package ca.recrown.islandsurvivalcraft.caching;

import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.ReadLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;

public class Cache<Key, Value> {
    private final int maxSize;
    private final ConcurrentHashMap<Key, CacheValue<Value>> data;
    private final ConcurrentLinkedQueue<Key> occurrenceOrder;
    private ReentrantReadWriteLock cleaningLock = new ReentrantReadWriteLock(true);
    private ReadLock writeLock = cleaningLock.readLock();
    private WriteLock readLock = cleaningLock.writeLock();

    public Cache(int maxSize) {
        data = new ConcurrentHashMap<>(maxSize + 1, 0.75f, 6);
        occurrenceOrder = new ConcurrentLinkedQueue<>();
        this.maxSize = maxSize;
    }

    public Cache() {
        this(1024);
    }

    public void setValue(Key key, Value value) {
        CacheValue<Value> previous = data.get(key);
        CacheValue<Value> fresh = new CacheValue<>(value);
        if (previous != null) fresh.occurrence += previous.occurrence;
        occurrenceOrder.add(key);
        data.put(key, fresh);

        if (data.size() > maxSize) {
            int occ = 0;
            do {
                Key potentialKey = occurrenceOrder.poll();
                CacheValue<Value> potential = data.get(potentialKey);
                potential.occurrence--;
                writeLock.lock();
                try {
                    occ = potential.occurrence;
                    if (occ < 1) {
                        data.remove(potentialKey);
                    } 
                } finally {
                    writeLock.unlock();
                }
            } while (occ > 0);
        }
    }

    /**
     * Retrieves cached value associated to the key.
     * Uses equals.
     * If does not exist, returns null.
     * @param key the key associated with the value.
     * @return the value associated to the key.
     */
    public Value getValue(Key key) {
        CacheValue<Value> res = null;
        readLock.lock();
        try {
            res = data.get(key);
            if (res == null) return null;
            res.occurrence++;
            occurrenceOrder.add(key);
        } finally {
            readLock.unlock();
        }
        return res.data;
    }

    public void clearCache() {
        data.clear();
        occurrenceOrder.clear();
    }
}