package ca.recrown.islandsurvivalcraft.world.generation;

import org.bukkit.Chunk;
import org.bukkit.World;
import org.bukkit.block.Biome;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.world.ChunkLoadEvent;

import ca.recrown.islandsurvivalcraft.Utilities;
import ca.recrown.islandsurvivalcraft.Types.Point2;
import ca.recrown.islandsurvivalcraft.caching.Cache;
import ca.recrown.islandsurvivalcraft.pathfinding.CoordinateTargetValidatable;
import ca.recrown.islandsurvivalcraft.pathfinding.CoordinateValidatable;
import ca.recrown.islandsurvivalcraft.pathfinding.DepthFirstSearch;
import ca.recrown.islandsurvivalcraft.world.BiomeSelector;
import ca.recrown.islandsurvivalcraft.world.IslandWorldMapper;

//Note: technically, the validators have to be run on land, and so, some condition checks may not be nessecary.
public class BiomePerIslandGenerator implements BiomeGenerator {
    private volatile boolean initialized;
    private final TemperatureMapGenerator temperatureMapGenerator;
    private final Cache<Point2, Biome> chunkBiomesCache;
    private final Cache<Point2, Boolean> chunkGenStatusCache;
    private volatile IslandWorldMapper worldIslandMap;
    private volatile BiomeSelector biomeSelector;
    private volatile World world;
    
    public BiomePerIslandGenerator() {
        this.temperatureMapGenerator = new TemperatureMapGenerator();
        chunkBiomesCache = new Cache<>(65536);
        chunkGenStatusCache = new Cache<>(65536);
    }
    
    @Override
    public void initialize(final World world, final IslandWorldMapper mapGenerator, final BiomeSelector biomeSelector) {
        if (initialized) throw new IllegalStateException("Biome generator already initialized.");
        initialized = true;
        this.world = world;
        this.worldIslandMap = mapGenerator;
        this.biomeSelector = biomeSelector;
        this.temperatureMapGenerator.setSeed(world.getSeed());
    }

    @EventHandler(priority = EventPriority.MONITOR)
    public void onChunkLoaded(ChunkLoadEvent event) {
        if (event.isNewChunk()) {
            Chunk chunk = event.getChunk();
            if (chunk.getWorld().getUID().hashCode() == world.getUID().hashCode() && chunk.getWorld().getUID().equals(world.getUID())) {
                chunkGenStatusCache.setValue(new Point2(chunk.getX(), chunk.getZ()), true);
            }
        }
    }

    @Override
    public BiomeGenerator getInstance() {
        return new BiomePerIslandGenerator();
    }

    @Override
    public Biome GenerateBiome(int chunkX, int chunkZ, int localX, int localZ) {
        Point2 currChunkCoords = new Point2(chunkX, chunkZ);
    
        int worldX = 16 * chunkX + localX;
        int worldZ = 16 * chunkZ + localZ;
        
        Point2 worldCoords = new Point2(worldX, worldZ);

        Biome cachedBiome = getStoredBiome(worldCoords);
        if (cachedBiome != null) return cachedBiome;
        
        float temperature = temperatureMapGenerator.getTemperature(worldX, worldZ);
        if (!worldIslandMap.isIsland(worldX, worldZ)) {
            return biomeSelector.getOceanBiome(temperature);
        }

        IslandBiomeInfo biomeInfo = new IslandBiomeInfo();
        DepthFirstSearch prevBiomeFinder = new DepthFirstSearch(biomeInfo);
        prevBiomeFinder.setStartPosition(worldX, worldZ);
        if (!prevBiomeFinder.findTarget(biomeInfo)) {
            if (biomeInfo.main == null) biomeInfo.main = biomeSelector.getLandBiome(temperature);
            if (biomeInfo.shore == null) biomeInfo.shore = biomeSelector.getShoreBiome(biomeInfo.main, temperature);
            if (biomeInfo.shallow == null) biomeInfo.shallow = biomeSelector.getOceanBiome(temperature);            
        }
        PropagatorInfo propInfo = new PropagatorInfo(biomeInfo.main, biomeInfo.shore, biomeInfo.shallow, currChunkCoords);
        DepthFirstSearch prop = new DepthFirstSearch(propInfo);
        prop.setStartPosition(worldX, worldZ);
        prop.findTarget(propInfo);
        
        if (worldIslandMap.isLand(worldX, worldZ)) {
            if (worldIslandMap.isShore(worldX, worldZ)) {
                return biomeInfo.shore;
            } else {
                return biomeInfo.main;
            }
        } else {
            return biomeInfo.shallow;
        }
    }

    private Biome getStoredBiome(Point2 worldCoords) {
        Point2 chunkCoords = Utilities.worldToChunkCoordinates(worldCoords);

        //Search our cache first.
        Biome biome = chunkBiomesCache.getValue(worldCoords);
        if (biome != null) return biome;

        //Check to see if bukkit has anything to say about it.
        Boolean chunkGenStat = chunkGenStatusCache.getValue(chunkCoords);
        if (chunkGenStat == null) {
            chunkGenStat = world.isChunkGenerated(chunkCoords.x, chunkCoords.y);
            chunkGenStatusCache.setValue(chunkCoords, chunkGenStat);
        }
        
        if (chunkGenStat) {
            biome = world.getBiome(worldCoords.x, 0, worldCoords.y);
            chunkBiomesCache.setValue(worldCoords, biome);
            return biome;
        }

        //Guess it's not cached or created.
        return null;
    }

    private void setCacheBiome(Point2 worldCoords, Biome biome) {
        //Set it in cache.
        chunkBiomesCache.setValue(worldCoords, biome);
    }

    private class PropagatorInfo implements CoordinateTargetValidatable, CoordinateValidatable {
        public final Biome main, shore, shallow;
        private final Point2 currentChunkCoords;
        private boolean hasSet;

        @Override
        public boolean isCoordinateTarget(int x, int y) {
            hasSet = false;
            Point2 worldCoords = new Point2(x, y);
            if (worldIslandMap.isLand(x, y)) {
                if (worldIslandMap.isShore(x, y)) {
                    setCacheBiome(worldCoords, shore);
                    hasSet = true;
                } else {
                    setCacheBiome(worldCoords, main);
                    hasSet = true;
                }
            } else {
                setCacheBiome(worldCoords, shallow);
                hasSet = true;
            }
            if (!hasSet) throw new IllegalStateException();
            return false;
        }
        
        @Override
        public boolean validate(int x, int y) {
            Point2 chunkCoords = Utilities.worldToChunkCoordinates(x, y);
            return chunkCoords.fastEquals(currentChunkCoords) && worldIslandMap.isIsland(x, y);
        }

        public PropagatorInfo(Biome main, Biome shore, Biome shallow, Point2 currentChunkCoords) {
            this.main = main;
            this.shore = shore;
            this.shallow = shallow;
            this.currentChunkCoords = currentChunkCoords;
        }
    }

    private class IslandBiomeInfo implements CoordinateTargetValidatable, CoordinateValidatable {
        private boolean isLand, isShore, isShallow;
        public Biome main, shore, shallow;
        
        @Override
        public boolean validate(int x, int y) {
            isShore = false;
            isShallow = false;
            isLand = false;
            return worldIslandMap.isIsland(x, y);
        }

        @Override
        public boolean isCoordinateTarget(int x, int y) {
            Point2 worldCoords = new Point2(x, y);

            if ((isLand = worldIslandMap.isLand(x, y))) {
                isShore = worldIslandMap.isShore(x, y);
            } else {
                isShallow = worldIslandMap.isShallowPortion(x, y);
            }

            if (main == null && isLand && !isShore) {
                main = getStoredBiome(worldCoords);
            } else
            if (shore == null && isShore) {
                shore = getStoredBiome(worldCoords);
            } else
            if (shallow == null && isShallow) {
                shallow = getStoredBiome(worldCoords);
            }
            return allBiomesAcquired();
        }

        public boolean allBiomesAcquired() {
            return shore != null && main != null && shallow != null;
        }
    }

    
}